#define _PICOTERM
/*
Ansteuerung einer 4,3" TFT-Anzeige mit 480x272 und 6 Bit Farbdarstellung.
Ohne weitere Anpassung werden eingehende Zeichen in schwarz auf weissem Hintergrund
angezeigt. Dabei werden Steuerzeichen wie CR, LF, Clear, Home und Alarm beachtet. 
Am Bildende erfolgt ein automatischer 'scroll'.

Mit ESC-Steuersequenzen koennen Farbeinstellungen veraendert und Zeichen als
Grafik frei positioniert angezeigt werden. Daneben gibt es Funktionen fuer
Setzen eines Pixels, zeichen von Linie und Rechteck sowie das Fuellen einer
rechteckigen Flaeche.

Ein Farbwert besteht aus je zwei Bit fuer Rot, Gruen und Blau und den Attributen
BLINKEN und TRANSPARENT. Bei TRANSPARENT werden nur die aktiven Pixel in der 
Vordergrundfarbe beschrieben, der Hintergrund wird nicht veraendert.

Es existieren zwei Bildseiten, die mit gleichem Inhalt beschrieben und zyklisch mit
der Blinkfrequenz gewechselt werden. Wird die Schriftfarbe mit dem Attribut BLINKEN
versehen, wird die 2. Bildseite mit 'alt_color' beschrieben, was beim Seitenwechsel
den Blinkeffekt erzeugt.

2023-12-10
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#define VERSIONS_MELDUNG "mino PicoTerm-TFT 2023-12-10"
#include "system_tft_rp2040.h"
#include <dma_req.h>

uint32_t SystemCoreClock  = DEF_SYSCLOCK;   // 150 MHz, leicht uebertaktet
uint32_t F_xtal = MHZ_12;                   // onboard default
uint32_t ms_counter;                        // fuer freie Verwendung
extern void  init_dma_CH10_pio1();

volatile uint8_t ms_flag, adc_flag;         // ms-flags


void kurz_warten(void)                      // fuer kurze Verzoegerungen
{
  volatile int i = 10000;
  while(i--);
}  


// XOSC aktivieren, SystemCoreClock setzen und Taktsignale aktivieren
void init_clockx(void)
{
volatile int i;
  VREG_AND_CHIP_RESET-> VREG = 15<<VREG_AND_CHIP_RESET_VREG_VSEL_Pos | VREG_AND_CHIP_RESET_VREG_EN_Msk;
  kurz_warten();       // Vreg abwarten

  XOSC->CTRL     = (XOSC_CTRL_FREQ_RANGE_1_15MHZ << XOSC_CTRL_FREQ_RANGE_Pos);
  XOSC->STARTUP  = 100;             // ca. 2 ms bei 12 MHz
  XOSC_SET->CTRL = (XOSC_CTRL_ENABLE_ENABLE << XOSC_CTRL_ENABLE_Pos);
  while (!(XOSC->STATUS & XOSC_STATUS_STABLE_Msk));
  
  RESETS_CLR->RESET = RESETS_RESET_pll_sys_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pll_sys_Msk));

// PLL: 12 MHz * 100 = 1200 MHz
  PLL_SYS->CS = (1 << PLL_SYS_CS_REFDIV_Pos);
  PLL_SYS->FBDIV_INT = (SystemCoreClock*PLL_DIV1*PLL_DIV2)/F_xtal;  // typ. 600 MHz
// 1200 MHz / 8 = 150 MHz
  PLL_SYS->PRIM = (PLL_DIV1 << PLL_SYS_PRIM_POSTDIV1_Pos) | 
                  (PLL_DIV2 << PLL_SYS_PRIM_POSTDIV2_Pos);

  PLL_SYS_CLR->PWR = PLL_SYS_PWR_VCOPD_Msk | PLL_SYS_PWR_PD_Msk;
  while (!(PLL_SYS->CS & PLL_SYS_CS_LOCK_Msk));

  PLL_SYS_CLR->PWR = PLL_SYS_PWR_POSTDIVPD_Msk;

  CLOCKS->CLK_REF_CTRL = (CLOCKS_CLK_REF_CTRL_SRC_xosc_clksrc << CLOCKS_CLK_REF_CTRL_SRC_Pos);
  CLOCKS->CLK_SYS_CTRL = (CLOCKS_CLK_SYS_CTRL_AUXSRC_clksrc_pll_sys << CLOCKS_CLK_SYS_CTRL_AUXSRC_Pos);
  CLOCKS_SET->CLK_SYS_CTRL = (CLOCKS_CLK_SYS_CTRL_SRC_clksrc_clk_sys_aux << CLOCKS_CLK_SYS_CTRL_SRC_Pos);

  CLOCKS->CLK_PERI_CTRL = CLOCKS_CLK_PERI_CTRL_ENABLE_Msk |
      (CLOCKS_CLK_PERI_CTRL_AUXSRC_clk_sys << CLOCKS_CLK_PERI_CTRL_AUXSRC_Pos);

  CLOCKS->CLK_ADC_CTRL = CLOCKS_CLK_ADC_CTRL_ENABLE_Msk |
      (CLOCKS_CLK_ADC_CTRL_AUXSRC_xosc_clksrc << CLOCKS_CLK_ADC_CTRL_AUXSRC_Pos);

// enable GPIO
  RESETS_CLR->RESET = RESETS_RESET_io_bank0_Msk | 
                      RESETS_RESET_pads_bank0_Msk;
  while((RESETS->RESET_DONE & (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk)) !=
        (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk));
}

// XOSC aktivieren, SystemCoreClock setzen und Taktsignale aktivieren
void init_clock(void)
{
volatile int i;
  VREG_AND_CHIP_RESET_SET-> VREG = 15<<VREG_AND_CHIP_RESET_VREG_VSEL_Pos;
  while(!(VREG_AND_CHIP_RESET->VREG & VREG_AND_CHIP_RESET_VREG_ROK_Msk));
  kurz_warten();       // Vreg abwarten

  XOSC->CTRL     = (XOSC_CTRL_FREQ_RANGE_1_15MHZ << XOSC_CTRL_FREQ_RANGE_Pos);
  XOSC->STARTUP  = 100;             // ca. 2 ms bei 12 MHz
  XOSC_SET->CTRL = (XOSC_CTRL_ENABLE_ENABLE << XOSC_CTRL_ENABLE_Pos);
  while (!(XOSC->STATUS & XOSC_STATUS_STABLE_Msk));
  
  RESETS_CLR->RESET = RESETS_RESET_pll_sys_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pll_sys_Msk));

#define REF_DIV     2

  PLL_SYS->CS = (REF_DIV << PLL_SYS_CS_REFDIV_Pos); // PLL-Aufloesung vorgeben
  PLL_SYS->FBDIV_INT = (SystemCoreClock*PLL_DIV2*REF_DIV)/(F_xtal/PLL_DIV1);  // PLL-VCO typ. 1310 MHz
// eff. Taktfrequenz ermitteln  
  PLL_SYS->PRIM = (PLL_DIV1 << PLL_SYS_PRIM_POSTDIV1_Pos) | 
                  (PLL_DIV2 << PLL_SYS_PRIM_POSTDIV2_Pos);
  SystemCoreClock = PLL_SYS->FBDIV_INT *(F_xtal/REF_DIV) / 
                    (PLL_DIV1*PLL_DIV2);

  PLL_SYS_CLR->PWR = PLL_SYS_PWR_VCOPD_Msk | PLL_SYS_PWR_PD_Msk;
  while (!(PLL_SYS->CS & PLL_SYS_CS_LOCK_Msk));

  PLL_SYS_CLR->PWR = PLL_SYS_PWR_POSTDIVPD_Msk;

  CLOCKS->CLK_REF_CTRL = (CLOCKS_CLK_REF_CTRL_SRC_xosc_clksrc << CLOCKS_CLK_REF_CTRL_SRC_Pos);
  CLOCKS->CLK_SYS_CTRL = (CLOCKS_CLK_SYS_CTRL_AUXSRC_clksrc_pll_sys << CLOCKS_CLK_SYS_CTRL_AUXSRC_Pos);
  CLOCKS_SET->CLK_SYS_CTRL = (CLOCKS_CLK_SYS_CTRL_SRC_clksrc_clk_sys_aux << CLOCKS_CLK_SYS_CTRL_SRC_Pos);

  CLOCKS->CLK_PERI_CTRL = CLOCKS_CLK_PERI_CTRL_ENABLE_Msk |
      (CLOCKS_CLK_PERI_CTRL_AUXSRC_clk_sys << CLOCKS_CLK_PERI_CTRL_AUXSRC_Pos);

  CLOCKS->CLK_ADC_CTRL = CLOCKS_CLK_ADC_CTRL_ENABLE_Msk |
      (CLOCKS_CLK_ADC_CTRL_AUXSRC_xosc_clksrc << CLOCKS_CLK_ADC_CTRL_AUXSRC_Pos);

// enable GPIO
  RESETS_CLR->RESET = RESETS_RESET_io_bank0_Msk | 
                      RESETS_RESET_pads_bank0_Msk;
  while((RESETS->RESET_DONE & (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk)) !=
        (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk));
}
// alle Millisekunden aufrufen
void TIMER_IRQ_0_IRQHandler(void)
{
  TIMER->INTR = TIMER_INTR_ALARM_0_Msk;
  TIMER->ALARM0 = TIMER->TIMELR + MS_TEILER;
  ms_counter++;
  ms_flag = 1;
  adc_flag = 1;
}

// 64 bit Timer mit 1 us Takt und loeschen
void init_timer(void)
{
  RESETS_CLR->RESET = RESETS_RESET_timer_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_timer_Msk));

  NVIC_EnableIRQ(TIMER_IRQ_0_IRQn);
  NVIC_SetPriority(TIMER_IRQ_0_IRQn, PRIO_LOW);
  WATCHDOG->TICK = WATCHDOG_TICK_ENABLE_Msk | (F_xtal / MHZ_FAKTOR);
  TIMER->TIMELW = 0;
  TIMER->TIMEHW = 0;
  TIMER->ALARM0 = TIMER->TIMELR + MS_TEILER;
  TIMER->INTE = TIMER_INTE_ALARM_0_Msk;
  
}

void delay_ms(uint32_t ms)
{
  do {
    while(!ms_flag);
    ms_flag = 0;
  } while(--ms);
} 

// Startbildschirm kopieren, wenn vorhanden
void zeige_logo()
{
  static int n, i, startzeile, endzeile;
  uint8_t *p = (uint8_t *) (LOGO_ADRESSE+0x10);
  startzeile = *(uint32_t *) (LOGO_ADRESSE+0x08);
  endzeile = *(uint32_t *) (LOGO_ADRESSE+0x0c); 
  if(*(uint32_t *) LOGO_ADRESSE == 0x6f6e696d) {
    for(n = startzeile; n <= endzeile; n++) { 
      for(i = 0; i < SIZE_X; i++) {
        bildspeicher[0][n*SIZE_X+i] = *p;
        bildspeicher[1][n*SIZE_X+i] = *p++;
      }  
    } 
  }
}

void reset_io()
{
// reset diverse Peripherie beim programmstart
  RESETS_SET->RESET = RESETS_RESET_io_bank0_Msk | 
                      RESETS_RESET_pio0_Msk | 
                      RESETS_RESET_pio1_Msk | 
                      RESETS_RESET_dma_Msk | 
                      RESETS_RESET_pwm_Msk | 
                      RESETS_RESET_timer_Msk |                        
                      RESETS_RESET_pads_bank0_Msk;
  SIO->GPIO_OE_CLR = 0xffffffff;  // alle loeschen
  kurz_warten();
}

void farben()
{
short x1,x2,y1,y2;
char farbe, muster = 0;
  tft_clear();
  do {
    x1 = 0, x2 = MAX_X-1;
    y1 = 0, y2 = MAX_Y-1;
    for(farbe = 0; farbe < 64; farbe++) {
      zeichne_rechteck(x1,y1,x2,y2,farbe ^ muster);
      zeichne_rechteck(x1+1,y1+1,x2-1,y2-1,farbe ^ muster);
      zeichne_rechteck(x1+2,y1+2,x2-2,y2-2,farbe ^ muster);
      x1 += 3; x2 -= 3;
      y1 +=2; y2 -=2;
    }
    delay_ms(300);
    muster ++;
    muster &= 0x3f;   // ohne Attribute
  } while(1);   // auf Beruehrung vom touch warten
} 

int main()
{
  reset_io();
  init_clock();
  init_timer();
  tft_backlight(MIN_HELLIGKEIT);
  PADS_BANK0->GPIO9 = PADS_BANK0_GPIO0_PUE_Msk;     // Reset freigeben: pullup = 1
  
  init_dma_CH10_pio1();
  tft_clear();
  delay_ms(250);
  init_uart0();
  init_t7_touch();
  
	tft_font(2);
  tft_push_pop(PUSH_XY);
  tft_str_einmitten(VERSIONS_MELDUNG, 260, HELLGRAU, 1);
  tft_font(0);
  tft_push_pop(POP_XY);
	
  tft_backlight(MAX_HELLIGKEIT);
/*
  int m,n;
  for(m = 0; m < 8; m++) {
    for(n = 0; n < 32; n++) {
      tft_gchar((n)*(SCHRIFT_BREITE-0)*2,(m)*(SCHRIFT_HOEHE-0)*2, n+m*32);
    }
  }
  
 */ 
//  farben();
  hauptprogramm();
}


