#define _DMA_TFT

#include "system_tft_rp2040.h"
#include <dma_req.h>


extern volatile char blink_bit;

int dma_test;
void TFT_update(void);

#define DMA_BYTES 4     // immer vier auf einmal
#define DMA_32    2     // fr uint32 Werte

#define TFT_0     0
#define TFT_1     1
#define TFT_2     2
#define TFT_3     3
#define TFT_4     4
#define TFT_5     5
#define TFT_CLK   18 
#define TFT_DE    19 

#define BACKLIGHT_PIN 28
uint8_t bildspeicher[2][BILDGROESSE];

// TFT backlight per PWM_6A
#define PWM6_TOP  256
#define F_PWM6    5000              // 5 kHz
void tft_backlight(uint8_t temp)	  // 8 = max. brightness
{	
static int init=0;
const uint16_t b_tab[] = {3,6,12,25,40,70,120,170,PWM6_TOP-1};
  if(!init) {
    init = 1;
    RESETS_CLR->RESET = RESETS_RESET_pwm_Msk;
    while (!(RESETS->RESET_DONE & RESETS_RESET_pwm_Msk));
    IO_BANK0->GPIO28_CTRL = IO_BANK0_GPIO28_CTRL_FUNCSEL_pwm_a_6; // PWM-output
    PWM->CH6_TOP = PWM6_TOP;                                      // 100%
    PWM->CH6_DIV = SystemCoreClock/PWM6_TOP/F_PWM6 << PWM_CH6_DIV_INT_Pos;  // etwa 10 kHz
    PWM_SET->CH6_CSR = PWM_CH0_CSR_EN_Msk;
    SIO->GPIO_OE_SET = BIT(BACKLIGHT_PIN);
  }  
  if(temp >= MAX_HELLIGKEIT) temp = MAX_HELLIGKEIT;
	PWM->CH6_CC = b_tab[temp];
}

/*  function of frame generation using PIO1-SM0
    m = TVP               // VSYNC line count
    n = THP/TVP           // sync-loop count
    isr = pixel/line-1    // eff. pixel count

START_NEW_FRAME:
ADR0:   y = isr           // pixel-1 for DE-VSYNC
ADR1:   goto VSYNC        // start DE-VSYNC
NEW_LINE:
ADR2:   y = isr           // pixel-1 for one line
ADR3:   set CLK+DE,1
        do {
ADR4:     output data%8   // STOPPED if frame complete: IRQ0=1 and DMA-rdy=1
ADR5:     clr IRQ0        // avoid ISR
ADR6:     set CLK,0
ADR7:     set CLK,1
ADR8:   } while(y--)      // complete line
ADR9:   set clk,1 + DE,0  // start DE-HSYNC
ADR10:  y = n-1         
VSYNC:
        do {
ADR11:    x = m-1         // DE-HSYNC + DE-VSYNC loop
          do {            // DE = 0
ADR12:      set CLK,0
ADR13:      set CLK,1
ADR14:    } while(x--)
ADR15:  } while(y--)
ADR16:  set IRQ0,1        // DE-SYNC complete
        wrap NEW_LINE
*/
 
void init_PIO1_neu(void)
{
  uint32_t *prg;
// enable PIO1 
  RESETS_CLR->RESET = RESETS_RESET_pio1_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pio1_Msk));
  
// enable GPIOs
  while((RESETS->RESET_DONE & (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk)) !=
        (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk));

// 6 bit/pixel: TFT_D0 - TFT_D5
  IO_BANK0->GPIO0_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  IO_BANK0->GPIO1_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  IO_BANK0->GPIO2_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  IO_BANK0->GPIO3_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  IO_BANK0->GPIO4_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  IO_BANK0->GPIO5_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;
  IO_BANK0->GPIO18_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;   // TFT_CLK
  IO_BANK0->GPIO19_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio1_0;   // TFT_DE

  
// push-pull output
  SIO->GPIO_OE_SET = 0x3f;                    // TFT_D0 - TFT_D5
  SIO->GPIO_OE_SET = 1<<TFT_CLK | 1<<TFT_DE;  // CLK+DE
  
// set output group TFT_0 - TFT_3 starting @GPIO0
  PIO1->SM0_PINCTRL =  4<<PIO0_SM0_PINCTRL_SET_COUNT_Pos | 0<<PIO0_SM0_PINCTRL_SET_BASE_Pos; // |
  PIO1->SM0_INSTR = 0xe08f;                   // set pindirs TFT_D0 - TFT_D3
// set output group TFT_4 - TFT_5 starting @GPIO4 
  PIO1->SM0_PINCTRL =  2<<PIO0_SM0_PINCTRL_SET_COUNT_Pos | 4<<PIO0_SM0_PINCTRL_SET_BASE_Pos; // |
  PIO1->SM0_INSTR = 0xe083;                   // set pindirs TFT_D4 - TFT_D5
// set output mode: TFT_CLK + TFT_DE  
  PIO1->SM0_PINCTRL =  2 << PIO0_SM0_PINCTRL_SET_COUNT_Pos |      // signals TFT_CLK + TFT_DE
                        TFT_CLK<<PIO0_SM0_PINCTRL_SET_BASE_Pos |  // starting @TFT_CLK
                        6 << PIO0_SM0_PINCTRL_OUT_COUNT_Pos;      // OUT-cmd: TFT_0 - TFT_5 
  PIO1->SM0_INSTR = 0xe083;                   // set pindirs TFT_CLK + TFT_DE  
// init transfer mode
  PIO1->SM0_SHIFTCTRL |= PIO0_SM0_SHIFTCTRL_AUTOPULL_Msk |        // autopull of DMA-data
                          0 << PIO0_SM0_SHIFTCTRL_PULL_THRESH_Pos | // 32 bit
                        PIO0_SM0_SHIFTCTRL_FJOIN_TX_Msk;          // TX-Fifo = 8
// init isr = pixel count
  PIO1->TXF0 = SIZE_X-1;            // count: pixel / line
  PIO1->SM0_INSTR = 0x8080;         // pull: OSR = pixel count
  PIO1->SM0_INSTR = 0xa0c7;         // mov isr,osr save count
  
 
  prg = (uint32_t *)&PIO1->INSTR_MEM0;    // pgm_start
// ADR0 
  *prg++ = 0xa046;        // mov y, isr; load pixel count
// ADR1 
  *prg++ = 0x0000+11;     // start VSYNC
// ADR2 
// new line:  
  *prg++ = 0xa046;        // mov y, isr; load pixel count
// ADR3
  *prg++ = 0xe003;        // set CLK=1 DE=1
// ADR4  
// autopull 4 pixel  
  *prg++ = 0x6308;      // OUT pins,osr autopull + 3 x wait
// ADR5 
  *prg++ = 0xc040;      // clr IRQ0
  
// ADR6 
  *prg++ = 0xe602;      // set clk,0  6 x wait DE=1
// ADR7 
  *prg++ = 0xe103;      // set clk,1  1 x wait DE=1
// ADR8 
  *prg++ = 0x0080+4;    // jump y--, loop ADR4 whole line
// ADR9 
// starting HSYNC
  *prg++ = 0xe001;      // set CLK,1 DE=0
// ADR10 
  *prg++ = 0xe040+(THP/TVP)-1;    // set y,n-1 for HSYNC y = 10
// ADR11                // if VSYNC: y = pixel count
// loop_y:  
  *prg++ = 0xe020+TVP-1;// set x,m-1  m = 3
// ADR12
// loop_x:  
  *prg++ = 0xe600;      // set clk,0  6 x wait DE bleibt 0
// ADR13 
  *prg++ = 0xe501;      // set clk,1  5 x wait DE bleibt 0
// ADR14 
  *prg++ = 0x0040+12;   // jmp x--,loop_x
// ADR15 
  *prg++ = 0x0080+11;   // jmp y--,loop_y
// ADR16 
  *prg++ = 0xc000;      // set IRQ0
// ADR17 
//  *prg++ = 0x0002;      // jmp new line
  PIO1->SM0_EXECCTRL = (16 << PIO0_SM0_EXECCTRL_WRAP_TOP_Pos) | 
                        (2 << PIO0_SM0_EXECCTRL_WRAP_BOTTOM_Pos);
 
  PIO1->SM0_INSTR = 0x0000;         // jmp SM0 start
 
  PIO1->SM0_CLKDIV = 1<<PIO0_SM0_CLKDIV_INT_Pos;      // max. speed
  PIO1_SET->CTRL = (1 << (PIO0_CTRL_SM_ENABLE_Pos));  // start
}


void TFT_update(void)
{
  if((DMA->CH10_CTRL_TRIG & DMA_CH10_CTRL_TRIG_BUSY_Msk) ||
     !(PIO1->IRQ0_INTE & PIO0_INTR_SM0_Msk))
    return;
  PIO1->SM0_INSTR = 0x0000;         // start SM0 DE-VSYNC
// and trigger DMA
#ifdef MIT_BLINKEN
    DMA->CH10_AL3_READ_ADDR_TRIG = (uint32_t)bildspeicher[blink_bit];  // alternierender Bildinhalt
#else
    DMA->CH10_AL3_READ_ADDR_TRIG = (uint32_t)bildspeicher[0];          // normaler Bildinhalt
#endif
}

// DMA fuer Ausgabe eines ganzen Bildes initialisieren
void init_dma_CH10_pio1(void)
{
volatile int vsync_delay;  
  
  RESETS_CLR->RESET = RESETS_RESET_dma_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_dma_Msk));
  NVIC_EnableIRQ(DMA_IRQ_1_IRQn);
  NVIC_SetPriority(DMA_IRQ_1_IRQn, PRIO_LOW);       // geringee Prioritt
  NVIC_EnableIRQ(PIO1_IRQ_0_IRQn);
  NVIC_SetPriority(PIO1_IRQ_0_IRQn, PRIO_MID_LOW);  // geringe Prioritt

  DMA->INTS1 = BIT(DMA_KANAL_TFT);                  // clear flag CH10
  DMA_SET->INTE1 = BIT(DMA_KANAL_TFT);              // enable CH10-isr

// Startbedingung fr SM0  
  DMA->CH10_READ_ADDR = (uint32_t) &bildspeicher;
  DMA->CH10_WRITE_ADDR = (uint32_t) &PIO1->TXF0;
  DMA->CH10_TRANS_COUNT = SIZE_X*SIZE_Y/DMA_BYTES;              // gesamtes Bild ausgeben
  DMA->CH10_CTRL_TRIG =  DREQ_PIO1_TX0 << DMA_CH10_CTRL_TRIG_TREQ_SEL_Pos |     // SM0-TXF0
                        DMA_32 << DMA_CH10_CTRL_TRIG_DATA_SIZE_Pos |  // uint32_t Werte
                        10 << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |  // word
                        DMA_CH10_CTRL_TRIG_INCR_READ_Msk |      // Lesezeiger anpassen  
                        DMA_CH10_CTRL_TRIG_EN_Msk;              // enable
  init_PIO1_neu();
  while(DMA->CH10_CTRL_TRIG & DMA_CH10_CTRL_TRIG_BUSY_Msk);
  vsync_delay = SystemCoreClock/1000;
  while(vsync_delay--);           // 5 ms delay
  PIO1->SM0_INSTR = 0x0000; 
 
/*
// Testmuster schreiben:
  int m, n;
  for(n = 0; n < SIZE_X; n++) {
    for(m = 0; m < SIZE_Y; m++)
      bildspeicher[0][m*SIZE_X+n] = n/10;
  }
*/    
}

// PIO1_SM0_IRQ0 freigeben, wenn transfer count == 0
void DMA_IRQ_1_IRQHandler(void)
{
  // fuer Bildausgabe  
  if(DMA->INTS1 & BIT(DMA_KANAL_TFT)) {           // DMA-CH10 ?
    DMA->INTS1 = BIT(DMA_KANAL_TFT);              // DMA-CH10 loeschen
    PIO1_SET->IRQ0_INTE = PIO0_IRQ0_INTE_SM0_Msk; // enable IRQ0
  }
  __DSB();
}

// Bildausgabe neu starten ggf. mit alternierenden Startadressen
// DE-VSYNC anspringen und Transfer neuu triggern
void PIO1_IRQ_0_IRQHandler(void)
{
  if(PIO1->IRQ0_INTS & PIO0_IRQ0_INTS_SM0_Msk) {  // bei IRQ0 DE-VSYNC starten
    PIO1->SM0_INSTR = 0x0000;                     // jmp loopy: SM0 Vsync anspringen
    PIO1_CLR->IRQ0_INTE = PIO0_IRQ0_INTE_SM0_Msk; // disable IRQ0

#ifdef MIT_BLINKEN
    DMA->CH10_AL3_READ_ADDR_TRIG = (uint32_t)bildspeicher[blink_bit];  // alternierender Bildinhalt
#else
    DMA->CH10_AL3_READ_ADDR_TRIG = (uint32_t)bildspeicher[0];          // normaler Bildinhalt
#endif
  }
  __DSB();
}

