#define _TEXT_GRAFIK
/*
Alle Funktionen ohne Gewhr
Michael Nowak
2023-12-10
www.mino-elektronik.de
*/
#include <system_tft_rp2040.h>
#include "cp437-6x10-mod.dat"
#include "cp437-8x16.dat"
#include "cp850-8x16.dat"
#include "cp866-8x16.dat"

uint32_t baud_tab[] = {9600, 19200, 38400, 57600, 76800, 115200,
                       125000, 230400, 250000, 460800, 500000};
#define MAX_BAUD_TAB sizeof(baud_tab)/sizeof(baud_tab[0])

char temp_str[50];                              // fuer printf-verwendung

// referenzen vorab
void fuelle_rechteck(int32_t x1, int32_t y1, int32_t x2, int32_t y2, int8_t farbe);
void tft_font(uint8_t temp);

uint8_t *bild = (uint8_t *) &bildspeicher;      // eff. Bildanfang

// Konvertieren von CP850 auf teilweise Unicode-Zeichen
static uint8_t *aktuelle_schrift = (uint8_t *)cp850_8x16;   // default Zeichensatz
uint8_t uni_aktiv = 1; 
uint8_t akt_zeichensatz = 0;
uint8_t uni_cov_tab[128] =
  {
    0x80, 0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8a,0x8b,0x8c,0x8d,0x8e,0x8f, // 8
    0x90, 0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0x9b,0x9c,0x9d,0x9e,0x9f, // 9
    0xff, 0xad,0xbd,0x9c,0xcf,0xbe,0xdd,0xf5,0xf9,0xb8,0xa6,0xae,0xaa,0xf0,0xa9,0xee, // a
    0xf8, 0xf1,0xfd,0xfc,0xef,0xe6,0xf4,0xfa,0xf7,0xfb,0xa7,0xaf,0xac,0xab,0xf3,0xa8, // b
    0xb7, 0xb5,0xb6,0xc7,0x8e,0x8f,0x92,0x80,0xd4,0x90,0xd2,0xd3,0xde,0xd6,0xd7,0xd8, // c
    0xd1, 0xa5,0xe3,0xe0,0xe2,0xe5,0x99,0x9e,0x9d,0xeb,0xe9,0xea,0x9a,0xed,0xe8,0xe1, // d
    0x85, 0xa0,0x83,0xc6,0x84,0x86,0x91,0x87,0x8a,0x82,0x88,0x89,0x8d,0xa1,0x8c,0x8b, // e
    0xd0, 0xa4,0x95,0xa2,0x93,0xe4,0x94,0xf6,0x9b,0x97,0xa3,0x96,0x81,0xec,0xe7,0x98  // f
  };    
     
uint8_t zoom_x = MIN_ZOOM, zoom_y = MIN_ZOOM,                   // Init-Werte 
        hoehe = SCHRIFT_HOEHE, breite = SCHRIFT_BREITE;         // dto.

uint8_t f_color = SCHWARZ,          // Vordergrundfarbe
        b_color = WEISS,            // Hintergrundfarbe
        alt_color = SCHWARZ;        // alternative Farbe fr Blinken
int32_t akt_x=0, akt_y=0,           // Startadresse fr neues Zeichen
        x_raster, y_raster,         // Inkrement nach Zeichenausgabe
        zeile, spalte;              // fr Textmodus, Funktion noch ungenau
void lcd_zeile1(int8_t *s);         // Textausgabe in oberster Textzeile x/y-zoom 3/4

static uint8_t esc_flag = 0, befehl[20];    // fr esc_scanner
static uint8_t touchmeldung = 1;            // Ausgabe ber UART aktiv

// Stacks fr zoom, xy, zeichensatz und Farbe 
static uint8_t  zoom_stack_ptr, zoom_stack[MAX_ZOOM_STACK+1][4];
static uint16_t xy_stack_ptr, xy_stack[MAX_XY_STACK+1][2];
static uint16_t font_stack_ptr, font_stack[MAX_XY_STACK+1];
static uint8_t  color_stack_ptr, color_stack[MAX_COLOR_STACK+1][3];

void push_zoom(void)                // zoom_x und zoom_y als unit8_t auf Stack speichern
{
  if(zoom_stack_ptr < MAX_ZOOM_STACK) {
    zoom_stack[zoom_stack_ptr][0] = zoom_x;
    zoom_stack[zoom_stack_ptr][1] = zoom_y;
    zoom_stack[zoom_stack_ptr][2] = x_raster;
    zoom_stack[zoom_stack_ptr][3] = y_raster;
    zoom_stack_ptr++;
  }  
}

void pop_zoom(void)                 // zoom_x und zoom_y vom Stack holen
{
  if(zoom_stack_ptr) {
    zoom_stack_ptr--;
    zoom_x = zoom_stack[zoom_stack_ptr][0];
    zoom_y = zoom_stack[zoom_stack_ptr][1];
    x_raster = zoom_stack[zoom_stack_ptr][2];
    y_raster = zoom_stack[zoom_stack_ptr][3];
  }
}

void clear_zoom_stack()
{
  zoom_stack_ptr = 0;
}  

void push_xy(void)                  // akt_x und akt_y als unit16_t auf Stack speichern
{
  if(xy_stack_ptr < MAX_XY_STACK) {
    xy_stack[xy_stack_ptr][0] = akt_x;
    xy_stack[xy_stack_ptr][1] = akt_y;
    xy_stack_ptr++;
  }  
}

void pop_xy(void)                   // zoom_x und zoom_y vom Stack holen
{
  if(xy_stack_ptr) {
    xy_stack_ptr--;
    akt_x = xy_stack[xy_stack_ptr][0];
    akt_y = xy_stack[xy_stack_ptr][1];
  }  
}

void clear_xy_stack()
{
  xy_stack_ptr = 0;
}  
void push_font(void)                  // akt_x und akt_y als unit16_t auf Stack speichern
{
  if(font_stack_ptr < MAX_FONT_STACK) {
    font_stack[font_stack_ptr] = akt_zeichensatz;
    font_stack_ptr++;
  }  
}

void pop_font(void)                   // zoom_x und zoom_y vom Stack holen
{
  if(font_stack_ptr) {
    font_stack_ptr--;
    akt_zeichensatz = font_stack[font_stack_ptr];
    tft_font(akt_zeichensatz);
  }  
}

void clear_font_stack()
{
  font_stack_ptr = 0;
}  


void push_color(void)               // f_color, b_color und alternative Farbe auf Stack speichern
{
  if(color_stack_ptr < MAX_COLOR_STACK) {
    color_stack[color_stack_ptr][0] = f_color;
    color_stack[color_stack_ptr][1] = b_color;
    color_stack[color_stack_ptr][2] = alt_color;
    color_stack_ptr++;
  }  
}

void pop_color(void)                // vorherige Farben wiederherstellen
{
  if(color_stack_ptr) {
    color_stack_ptr--;
    f_color = color_stack[color_stack_ptr][0];
    b_color = color_stack[color_stack_ptr][1];
    alt_color = color_stack[color_stack_ptr][2];
  }  
}

void clear_color_stack()
{
  color_stack_ptr = 0;
}  

void set_pixel(uint32_t x, uint32_t y, uint8_t farbe)
{
uint8_t *adr = (uint8_t *) (bild + (x+y*BILDBREITE));
  
  if(x <= MAX_X && y <= MAX_Y) {
    *adr = farbe;
#ifdef MIT_BLINKEN                  // zustzlich alt. Bildspeicher beschreiben
    if(farbe&BLINKEN) *(adr+BILDGROESSE) = alt_color;
    else  *(adr+BILDGROESSE) = farbe;
#endif  
  }
}

void set_xy(uint16_t x, uint16_t y)
{
  akt_x = x;
  akt_y = y;
}

void schreibe_zeichen_dyn(uint8_t c, uint8_t farbe, uint8_t eff_hoehe)
{   
uint32_t zeile, spalte;             // anzahl Zeilen und Spalten pro Zeichen
uint32_t muster, i, j;
uint32_t eff_x, eff_y;
uint8_t *adr, *temp_adr;
uint8_t *zeichen;                   // Basisadresse des Zeichens

// Laufzeit testen
  IO_BANK0->GPIO14_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // Kontroll-Ausgang
  SIO->GPIO_OE_SET = BIT(14); 
//  SIO->GPIO_OUT_SET = BIT(14);

  if(uni_aktiv && (c >= 0x80))      // ggf. an Unicode anpassen
    c = uni_cov_tab[c&0x7f];
  zeichen = &aktuelle_schrift[(c+1)*ZEICHEN_HOEHE];  // Basisadresse des Zeichens
  
  eff_y = akt_y;
  if(eff_hoehe) zeile = eff_hoehe;  // Sollwert verwenden
  else zeile = ZEICHEN_HOEHE;       // sonst default
// testen, ob Zeichen voll in den Bildbereich passt
  if(akt_x >= 0 && akt_x + ZEICHEN_LAENGE * zoom_x < MAX_X && 
     akt_y >= 0 && akt_y + zeile * zoom_y < MAX_Y) {
// keine weiteren Tests beim Schreiben der Pixel notwendig       
    while(zeile--) {                // gesamte hhe
      eff_x = akt_x;
      muster = *zeichen++;          // daten der Zeile
      spalte = ZEICHEN_LAENGE;
      while(spalte--) {
        temp_adr = adr = (uint8_t *) (bild + (eff_x+eff_y*BILDBREITE));
// fr 2 x 2 optimieren        
        if(zoom_x == 2 && zoom_y == 2) {
           if(muster & 0x80) {
              if(!(b_color&TRANSPARENT)) {
                *adr = farbe;       // Pixel setzen
                *(adr+1) = farbe;   // 2 x Breite
                *(adr+BILDBREITE) = farbe;
                *(adr+BILDBREITE+1) = farbe;
#ifdef MIT_BLINKEN
                if(farbe&BLINKEN) { 
                  *(adr+BILDGROESSE) = alt_color;
                  *(adr+BILDGROESSE+1) = alt_color;
                  *(adr+BILDGROESSE+BILDBREITE) = alt_color;
                  *(adr+BILDGROESSE+BILDBREITE+1) = alt_color;
                } else   {
                  *(adr+BILDGROESSE) = farbe;
                  *(adr+BILDGROESSE+1) = farbe;
                  *(adr+BILDGROESSE+BILDBREITE) = farbe;
                  *(adr+BILDGROESSE+BILDBREITE+1) = farbe;
                }      
#endif
              }  
            } else {
              if(!(farbe&TRANSPARENT)) {
                *adr = b_color;       // b_color Pixel setzen
                *(adr+1) = b_color;   // 2 x Breite
                *(adr+BILDBREITE) = b_color;
                *(adr+BILDBREITE+1) = b_color;   // 2 x Hoehe
#ifdef MIT_BLINKEN
                *(adr+BILDGROESSE) = b_color;
                *(adr+BILDGROESSE+1) = b_color;
                *(adr+BILDBREITE+BILDGROESSE) = b_color;
                *(adr+BILDBREITE+BILDGROESSE+1) = b_color;
#endif
              }  
            }   
          } else 
// fr 1 x 1 optimieren 
          if(zoom_x == 1 && zoom_y == 1) {
            if(muster & 0x80) {
              if(!(b_color&TRANSPARENT)) {
                *adr = farbe;       // Pixel setzen
#ifdef MIT_BLINKEN
                if(farbe&BLINKEN) { 
                  *(adr+BILDGROESSE) = alt_color;
                } else   {
                  *(adr+BILDGROESSE) = farbe;
                }      
#endif
              }  
            } else {
              if(!(farbe&TRANSPARENT)) {
                *adr = b_color;     // Pixel setzen
#ifdef MIT_BLINKEN
                *(adr+BILDGROESSE) = b_color;
#endif                
              }  
            }   

          } else {
// Ohne Optimierung 1x1 oder 2x2            
          i= 0;
          while(i < zoom_y) {         // zoom y
            j = 0;
            while(j < zoom_x) {       // zoom x
              if(muster & 0x80) {
                if(!(b_color&TRANSPARENT)) {
                  *adr = farbe;       // Pixel setzen
#ifdef MIT_BLINKEN
                  if(farbe&BLINKEN) *(adr+BILDGROESSE) = alt_color;
                  else  *(adr+BILDGROESSE) = farbe;
#endif                
                }  
              } else {
                if(!(farbe&TRANSPARENT)) {
                  *adr = b_color;     // b_color Pixel setzen
#ifdef MIT_BLINKEN
                  *(adr+BILDGROESSE) = b_color;
#endif                
                }  
              }   
              adr++;
              j++;
            } // while(j)
            temp_adr += BILDBREITE;   
            adr = temp_adr;
            i++;
          } // while(i)
        }
        muster <<= 1;               // nchstes pixel
        eff_x += zoom_x;
      } // while(spalte)
      eff_y += zoom_y;              // nchste Zeile
    } // while(zeile)
  } else {                        
// Zeichen liegt am Bildrand; daher jedes Pixel auf gueltigen Bereich testen
    while(zeile--) {                // gesamte hhe
      eff_x = akt_x;
      muster = *zeichen++;          // daten der Zeile
      spalte = ZEICHEN_LAENGE;
      while(spalte--) {
        i= 0;
        while(i < zoom_y) {         // zoom y
          j = 0;
          while(j < zoom_x) {       // zoom x
            if(muster & 0x80) {
              if(!(b_color & TRANSPARENT)) {
                set_pixel(eff_x+j, eff_y+i,farbe);  // Pixel setzen
              }  
            } else {
              if(!(farbe & TRANSPARENT)) set_pixel(eff_x+j, eff_y+i,b_color); // b_color
            }   
            j++;
          } // while(j)
          i++;
        } // while(i)
        muster <<= 1;               // nchstes pixel
        eff_x += zoom_x;
      } // while(spalte)
      eff_y += zoom_y;              // nchste Zeile
    } // while(zeile)
  }    
  SIO->GPIO_OUT_CLR = BIT(14); 
}


void string_ausgabe(int8_t *s, uint8_t farbe)
{
uint8_t c;
  while((c = *s++)) {
    schreibe_zeichen_dyn(c,farbe, 0);
    akt_x += ZEICHEN_LAENGE * zoom_x; // Cursor anpassen
  }
}       

void schreibe_string(int8_t *s, uint8_t farbe)
{
  zoom_x = zoom_y = 1;
  string_ausgabe(s,farbe);
}

void schreibe_string2(int8_t *s, uint8_t farbe)
{
  zoom_x = zoom_y = 2;
  string_ausgabe(s,farbe);
}

void schreibe_string3(int8_t *s, uint8_t farbe)
{
  zoom_x = zoom_y = 3;
  string_ausgabe(s,farbe);
}


void schreibe_string_mittig(int8_t *s, uint8_t farbe)
{
uint8_t c;
  c=0; while(s[c]) c++;
  akt_x += (MAX_X - c * ZEICHEN_LAENGE*zoom_x)/2;    
  while((c = *s++)) {
    schreibe_zeichen_dyn(c,farbe, 0);
    akt_x += ZEICHEN_LAENGE*zoom_x;
  }     
}

// speziell fuer grosse Darstellung
void lcd_zeile_x(int8_t *s, uint8_t farbe)
{
uint8_t c;
  push_zoom();
  zoom_x = 3;zoom_y = 4;                          // SCHN GROSS
  c=0; while(s[c]) c++;                           // Zeichen zhlen
  akt_x = (MAX_X - c * ZEICHEN_LAENGE*zoom_x)/2;  // Startpunkt links ermitteln
  fuelle_rechteck(0, akt_y, akt_x, akt_y+ZEICHEN_HOEHE*zoom_y,WEISS); // links lschen
  while((c = *s++)) {
    schreibe_zeichen_dyn(c,farbe, 0);
  akt_x += ZEICHEN_LAENGE*zoom_x;
  }     
  fuelle_rechteck(akt_x, akt_y, MAX_X, akt_y+ZEICHEN_HOEHE*zoom_y,WEISS); // rechts lschen
  pop_zoom();
}

void tft_str_einmitten(int8_t *s, uint16_t y_pos, uint8_t farbe, int8_t zoom)
{
uint16_t temp_y;
  temp_y = akt_y;
  push_zoom();
  zoom_x = zoom_y = zoom;
  akt_y = y_pos;
  schreibe_string_mittig(s, farbe);
  pop_zoom();
  akt_y = temp_y;
}

void DMA11_transfer_abwarten(void)
{
  while(DMA->CH11_CTRL_TRIG & DMA_CH11_CTRL_TRIG_BUSY_Msk);
}

// optimiert mit DMA
void horizontale_linie_DMA(int32_t x1, int32_t x2, int32_t y, uint8_t farbe)
{
uint16_t temp;
uint8_t *adr;   
  if(x1 < 0)  x1 = 0;
  if(x2 < 0)  x2 = 0;
  if(y < 0)   y = 0;
  if(x1 > MAX_X)  x1 = MAX_X;
  if(x2 > MAX_X)  x2 = MAX_X;
  if(y > MAX_Y)   y = MAX_Y;

  if(x1 > x2) {
    temp = x1;
    x1 = x2;
    x2 = temp;
  } 
  adr = (bild + (x1+y*BILDBREITE));
  DMA11_transfer_abwarten();  

  DMA->CH11_READ_ADDR = (uint32_t) &farbe;
  DMA->CH11_WRITE_ADDR = (uint32_t) adr;
  DMA->CH11_TRANS_COUNT = x2-x1+1;           // 
  DMA->CH11_CTRL_TRIG = 0x3f<<DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos |
    0 << DMA_CH11_CTRL_TRIG_DATA_SIZE_Pos |  // byteweise
                        DMA_KANAL_LINIE << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |  // keine Verkettung
                        DMA_CH11_CTRL_TRIG_INCR_WRITE_Msk |       // Schreibzeiger anpassen  
                        DMA_CH11_CTRL_TRIG_EN_Msk;               // enable
  DMA11_transfer_abwarten();  
  
#ifdef MIT_BLINKEN          // Hintergrundfarbe hinzufgen    
    if(!(farbe&BLINKEN))
      DMA->CH11_READ_ADDR = (uint32_t) &farbe;
    else
      DMA->CH11_READ_ADDR = (uint32_t) &alt_color;
  DMA->CH11_WRITE_ADDR = (uint32_t) (adr+BILDGROESSE);
  DMA->CH11_TRANS_COUNT = x2-x1+1;           // 
  DMA->CH11_CTRL_TRIG = 0x3f<<DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos |
                        0 << DMA_CH11_CTRL_TRIG_DATA_SIZE_Pos |  // byteweise
                        DMA_KANAL_LINIE << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |  // keine Verkettung
                        DMA_CH11_CTRL_TRIG_INCR_WRITE_Msk |       // Schreibzeiger anpassen  
                        DMA_CH11_CTRL_TRIG_EN_Msk;               // enable
#endif      
}

// langsam da ohne DMA
void horizontale_linie(int32_t x1, int32_t x2, int32_t y, uint8_t farbe)
{
uint16_t temp;
uint8_t *adr;   
  if(x1 < 0)  x1 = 0;
  if(x2 < 0)  x2 = 0;
  if(y < 0)   y = 0;
  if(x1 > MAX_X)  x1 = MAX_X;
  if(x2 > MAX_X)  x2 = MAX_X;
  if(y > MAX_Y)   y = MAX_Y;

  if(x1 > x2) {
    temp = x1;
    x1 = x2;
    x2 = temp;
  } 
#ifdef MIT_BLINKEN      
    if(!(farbe&BLINKEN)) temp = farbe;
    else temp = alt_color;
#endif      
  adr = (bild + (x1+y*BILDBREITE));
  while(x1 <= x2) {
    *adr = farbe;
#ifdef MIT_BLINKEN      
    *(adr + BILDGROESSE) = temp;  // Hintergrundfarbe
#endif      
    adr++;
    x1++;
  }
}

// optimiert fuer senkrechte Linie
void vertikale_linie(int32_t x, int32_t y1, int32_t y2, uint8_t farbe)
{
uint16_t temp;
uint8_t *adr;   
  if(y1 < 0)  y1 = 0;
  if(y2 < 0)  y2 = 0;
  if(x < 0)   x = 0;
  if(y1 > MAX_Y)  y1 = MAX_Y;
  if(y2 > MAX_Y)  y2 = MAX_Y;
  if(x > MAX_X)   x = MAX_X;

  if(y1 > y2) {
    temp = y1;
    y1 = y2;
    y2 = temp;
  } 
#ifdef MIT_BLINKEN      
    if(!(farbe&BLINKEN)) temp = farbe;
    else temp = alt_color;
#endif      
  adr = (bild + (x+y1*BILDBREITE));
  while(y1 <= y2) {
    *adr = farbe;
#ifdef MIT_BLINKEN      
    *(adr + BILDGROESSE) = temp;  // Hintergrundfarbe
#endif      
    adr += BILDBREITE;
    y1++;
  }
}

int sgn(int x){
  return (x > 0) ? 1 : (x < 0) ? -1 : 0;
}

// fuer beliebige Linien, ggf. optimiert
void zeichne_linie(int x1, int y1, int x2, int y2, int8_t farbe)
{
int x, y, t, dx, dy, incx, incy, pdx, pdy, ddx, ddy, es, el, err;
// auf Optimierung testen
  if(y1 == y2)                          // horizontale Linie
    horizontale_linie_DMA(x1, y1, y2, farbe);
  else if(x1 == x2)                     // vertikale Linie
    vertikale_linie(x1, y1, y2, farbe);
  else if(x1 == x2 && y1 == y2)         // nur ein Pixel
    set_pixel(x1,y1, farbe);
  else { 
    dx = x2 - x1;
    dy = y2 - y1;
    incx = sgn(dx);
    incy = sgn(dy);
    if(dx<0) dx = -dx;
    if(dy<0) dy = -dy;

    if (dx>dy){
      pdx=incx; pdy=0;
      ddx=incx; ddy=incy;
      es =dy;   el =dx;
    } else {
      pdx=0;    pdy=incy;
      ddx=incx; ddy=incy;
      es =dx;   el =dy;
    }

    x = x1;
    y = y1;
    err = el/2;
    set_pixel(x,y, farbe);

    for(t=0; t<el; ++t) {
      err -= es;
      if(err<0) {
        err += el;
        x += ddx;
        y += ddy;
      } else {
        x += pdx;
      y += pdy;
      }
      set_pixel(x,y, farbe);
    }
  }  
}

// nur Umriss
void zeichne_rechteck(int32_t x1, int32_t y1, int32_t x2, int32_t y2, int8_t farbe)
{
  vertikale_linie(x1, y1, y2, farbe);       // links senkrecht
  vertikale_linie(x2, y1, y2, farbe);       // rechts senkrecht
  horizontale_linie_DMA(x1, x2, y1, farbe); // oben
  horizontale_linie_DMA(x1, x2, y2, farbe); // unten
  DMA11_transfer_abwarten();  
}

// ausgefuelltes Feld
void fuelle_rechteck(int32_t x1, int32_t y1, int32_t x2, int32_t y2, int8_t farbe)
{
int16_t temp;   
  if(y1 > y2) {
    temp = y1;
    y1 = y2;
    y2 = temp;
  }
  while(y1 <= y2) {
    horizontale_linie_DMA(x1, x2, y1, farbe);
    y1++;
  }
  DMA11_transfer_abwarten();  
}   

void loesche_tft(uint8_t farbe)
{
  fuelle_rechteck(0,0,MAX_X,MAX_Y,farbe);
}

void schreibe_zeichen(unsigned char c, char farbe)
{
//  zoom_y = zoom_x;    // vormals
  schreibe_zeichen_dyn(c, farbe, 0);
}   


void tft_clear(void)
{
  fuelle_rechteck(0,0,MAX_X,MAX_Y,(b_color)); // Hintergrundfarbe
  akt_x = 0;
  akt_y = 0;
  zeile = 0;
  spalte = 0;
  x_raster = breite * zoom_x;
  y_raster = hoehe * zoom_y;
}

// org. Funktion ohne DMA
void copy_line_org(int dest, int src)
{
static uint8_t *adr_dest, *adr_src;
int n;  
    if(dest > MAX_Y || src > MAX_Y) return;
    if(dest < 0 || src < 0) return;
    adr_src = (bild + (src * BILDBREITE));
    adr_dest = (bild + (dest * BILDBREITE));
    for(n = 0; n < H_LINE_LEN; n++)
      *(uint8_t *)adr_dest++ = *(uint8_t *)adr_src++;
#ifdef MIT_BLINKEN
    adr_src = (bild + (src * BILDBREITE) + BILDGROESSE);
    adr_dest = (bild + (dest * BILDBREITE) + BILDGROESSE);
    for(n = 0; n < H_LINE_LEN; n++)
      *(uint8_t *)adr_dest++ = *(uint8_t *)adr_src++;

#endif    
}

// schnell per DMA
void copy_line(int dest, int src)
{
static uint8_t *adr_dest, *adr_src;
  
  if(dest > MAX_Y || src > MAX_Y) return;
  if(dest < 0 || src < 0) return;
  adr_src = (bild + (src * BILDBREITE));
  adr_dest = (bild + (dest * BILDBREITE));
  DMA11_transfer_abwarten();  

  DMA->CH11_READ_ADDR = (uint32_t) adr_src;
  DMA->CH11_WRITE_ADDR = (uint32_t) adr_dest;
  DMA->CH11_TRANS_COUNT = H_LINE_LEN/4;           // geht geht schneller
  DMA->CH11_CTRL_TRIG = 0x3f<<DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos |
                        2 << DMA_CH11_CTRL_TRIG_DATA_SIZE_Pos |   // immer vier Bytes
                        DMA_KANAL_LINIE << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |  // keine Verkettung
                        DMA_CH11_CTRL_TRIG_INCR_WRITE_Msk |       // Schreibzeiger anpassen  
                        DMA_CH11_CTRL_TRIG_INCR_READ_Msk |        // Schreibzeiger anpassen  
                        DMA_CH11_CTRL_TRIG_EN_Msk;                // enable
#ifdef MIT_BLINKEN
  adr_src = (bild + (src * BILDBREITE) + BILDGROESSE);
  adr_dest = (bild + (dest * BILDBREITE) + BILDGROESSE);
  DMA11_transfer_abwarten();  
  DMA->CH11_READ_ADDR = (uint32_t) adr_src;
  DMA->CH11_WRITE_ADDR = (uint32_t) adr_dest;
  DMA->CH11_TRANS_COUNT = H_LINE_LEN/4;           // geht geht schneller
  DMA->CH11_CTRL_TRIG = 0x3f<<DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos |
                        2 << DMA_CH11_CTRL_TRIG_DATA_SIZE_Pos |   // immer vier Bytes
                        DMA_KANAL_LINIE << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |  // keine Verkettung
                        DMA_CH11_CTRL_TRIG_INCR_WRITE_Msk |       // Schreibzeiger anpassen  
                        DMA_CH11_CTRL_TRIG_INCR_READ_Msk |        // Schreibzeiger anpassen  
                        DMA_CH11_CTRL_TRIG_EN_Msk;                // enable
#endif    
}

// Zeile fuer Zeile per DMA
void scroll(void)
{
static short y0, y_diff;
uint32_t temp_col;
  SIO->GPIO_OUT_SET = BIT(14);      // Zeit messen
  y_diff = y_raster;
  for(y0 = 0; y0 <= MAX_Y-y_diff; y0++) {
    copy_line(y0, y0+y_diff);         // per DMA
  }
  temp_col = b_color<<24 | b_color<<16 | b_color<<8 | b_color;
  DMA11_transfer_abwarten();  

// freie Zeile unten loeschen  
  DMA->CH11_READ_ADDR = (uint32_t) &temp_col;
  DMA->CH11_WRITE_ADDR = (uint32_t) bild+y0*BILDBREITE;
  DMA->CH11_TRANS_COUNT = (SIZE_Y - y0) * H_LINE_LEN/4;           // geht geht schneller
  DMA->CH11_CTRL_TRIG = 0x3f<<DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos |
                        2 << DMA_CH11_CTRL_TRIG_DATA_SIZE_Pos |   // immer vier Bytes
                        DMA_KANAL_LINIE << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |  // keine Verkettung
                        DMA_CH11_CTRL_TRIG_INCR_WRITE_Msk |       // Schreibzeiger anpassen  
                        DMA_CH11_CTRL_TRIG_EN_Msk;                // enable
#ifdef MIT_BLINKEN
  DMA11_transfer_abwarten();  
  DMA->CH11_READ_ADDR = (uint32_t) &temp_col;
  DMA->CH11_WRITE_ADDR = (uint32_t) bild+y0*BILDBREITE+BILDGROESSE;
  DMA->CH11_TRANS_COUNT = (SIZE_Y - y0) * H_LINE_LEN/4;           // geht geht schneller
  DMA->CH11_CTRL_TRIG = 0x3f<<DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos |
                        2 << DMA_CH11_CTRL_TRIG_DATA_SIZE_Pos |   // immer vier Bytes
                        DMA_KANAL_LINIE << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |  // keine Verkettung
                        DMA_CH11_CTRL_TRIG_INCR_WRITE_Msk |       // Schreibzeiger anpassen  
                        DMA_CH11_CTRL_TRIG_EN_Msk;                // enable
#endif
  DMA11_transfer_abwarten();  
  SIO->GPIO_OUT_CLR = BIT(14);      // Zeit messen
}

// zeichensatz waehlen
void tft_font(uint8_t temp)
{
  uni_aktiv = 0;                    // default keine Konvertierung
  akt_zeichensatz = temp;
  switch(temp) {
    default:
    case 0:
      uni_aktiv = 1;                // mit teilw. Konvertierung auf uni-Zeichensatz
      aktuelle_schrift = (uint8_t *)cp850_8x16;
      break;
    case 1:
      aktuelle_schrift = (uint8_t *)cp850_8x16;
      break;
    case 2:
		  aktuelle_schrift = (uint8_t *)cp437_6x10;
      break;
    case 3:
      aktuelle_schrift = (uint8_t *)cp437_8x16;
      break;
    case 4:
      aktuelle_schrift = (uint8_t *)cp866_8x16;
      break;

	} 
  hoehe = ZEICHEN_HOEHE;
  breite = ZEICHEN_LAENGE;
	x_raster = breite * zoom_x;
	y_raster = hoehe * zoom_y;
}	  

// Ausgabe fuer Text, Cursor anpassen und auto-scroll am Bildende
void tft_ausgabe(char c)                 
{
  schreibe_zeichen(c, f_color);
  akt_x += x_raster;
  if(akt_x > MAX_X) {
    akt_x = 0;
    akt_y += y_raster;
    if(akt_y + y_raster > MAX_Y+1) {
      scroll();
      akt_y -= y_raster;
    }    
  }
}

void set_zoom_xy(int x, int y)
{
  if(x < MIN_ZOOM) x = MIN_ZOOM;
  if(x > MAX_ZOOM) x = MAX_ZOOM;
  if(y < MIN_ZOOM) y = MIN_ZOOM;
  if(y > MAX_ZOOM) y = MAX_ZOOM;
  zoom_x = x;
  zoom_y = y;
  x_raster = breite * zoom_x;
  y_raster = hoehe * zoom_y;
} 

// y = x, wenn y == 0
void tft_zoom(uint8_t zoom)     // oberes nibble = y, unteres = x
{
uint8_t temp_x, temp_y;  
  temp_x = zoom & 0xf;          // unteres nibble zoom_x,
  temp_y = (zoom >> 4) & 0xf;   // oberes nibble zoom_y,
  if(!temp_y) temp_y = temp_x;  // falls == 0, dann x-Wert verwenden
  set_zoom_xy(temp_x, temp_y);
}  
      
// wenn erlaubt, Koordinaten und touch-Status ausgeben
static void sende_touch()
{
static uint8_t temp;
int t_x, t_y, t_aktiv;
    lese_touch(&t_x, &t_y);
    t_aktiv = touch_aktiv;
    touch_aktiv = 0;                // nur einmal ausgeben
    if(!touchmeldung) return;       // nicht zulaessig
    
   
    temp = ((t_x >> 7) & 0x3f) | 0x80;
    if(t_aktiv == 1) temp |= 0x40;  // Kennung fr tastendruck
    uart0_putc(temp);               // obere 5 bits von x
    temp = (t_x & 0x7f);            // restliche 7 bit
    uart0_putc(temp);               // obere 5 bits von x
    temp = ((t_y >> 7) & 0x3f);     // obere 5 bits von y
    uart0_putc(temp);               // obere 5 bits von x
    temp = (t_y & 0x7f);            // restliche 7 bit
    uart0_putc(temp);               // obere 5 bits von x
}   


void tft_putstr(char *p)
{ 
  while(*p) tft_ausgabe(*p++);
}

void tft_putstrn(char *p, char n)
{
  while(n--) tft_ausgabe(*p++);
}

void tft_gchar(short x1,short y1, uint8_t z)
{
  push_xy();        // Cursor fr Zeichenausgabe nicht verndern
  akt_x = x1;
  akt_y = y1;
  schreibe_zeichen(z, f_color);
  pop_xy();
}


void tft_set_gcursor(short x1,short y1, uint8_t abstand)
{
  akt_x = x1;
  akt_y = y1;
  x_raster = abstand;
}

void tft_gtext(short x, short y, short xraster , char *p)
{
  while(*p){
    tft_gchar(x, y, *p++);
    x += xraster;
  }
}

void grafik_xy_pos(int x_pos, int y_pos)
{
  if(x_pos < 0) y_pos = 0;
  if(y_pos < 0) y_pos = 0;
  x_raster = breite * zoom_x;
  y_raster = hoehe * zoom_y;
  akt_x = x_pos;
  akt_y = y_pos;
}

void tft_set_textcursor(uint8_t z, uint8_t s)
{
  zeile = z-1;                      // logisch 1,1 ist oben links 0,0
  spalte = s-1;
  if(zeile < 0) zeile = 0;
  if(spalte < 0) spalte = 0;
  x_raster = breite * zoom_x;
  y_raster = hoehe * zoom_y;
  akt_x = spalte * x_raster;
  if(akt_x > MAX_X) akt_x = 0;      // bei Fehler auf 0 setzen
  akt_y = zeile * y_raster;
  if(akt_y > MAX_Y) akt_y = 0;      // bei Fehler auf 0 setzen
}  

void blanks_entfernen(char *s)
{
char temp,i;
  temp = strlen(s)-1;
  while(temp && (s[temp] == ' ')) s[temp--] = 0; /* blanks am ende loeschen */
  temp = 0;
  while(s[temp] == ' ') temp++;     /* blanks am anfang suchen */
  if(temp) {                        /* nur, wenn welche vorhanden */
    i=0;
    do {
      s[i++] = s[temp];             /* verschieben */
    } while(s[temp++]);             /* bis ende erreicht */
  }
}

void tft_gr_cursor_on(short offset, char on_farbe)
{
  fuelle_rechteck(akt_x+offset,
        akt_y+SCHRIFT_HOEHE*zoom_y-2,
        akt_x+offset+SCHRIFT_BREITE*zoom_x,
        akt_y+SCHRIFT_HOEHE*zoom_y-3,
        on_farbe);
}

void tft_gr_cursor_off(short offset, char off_farbe)
{
  fuelle_rechteck(akt_x+offset,
        akt_y+SCHRIFT_HOEHE*zoom_y-2,
        akt_x+offset+SCHRIFT_BREITE*zoom_x,
        akt_y+SCHRIFT_HOEHE*zoom_y-3,
        off_farbe);
}

void tft_bd_index(uint8_t bd_index)
{
  if(bd_index > MAX_BAUD_TAB)
    set_baudrate0(DEF_BAUDRATE);
  else
    set_baudrate0(baud_tab[bd_index]);
}  

void tft_fcol(uint8_t farbe)
{ f_color = farbe;}

void tft_bcol(uint8_t farbe)
{ b_color = farbe;}

void tft_acol(uint8_t farbe)
{ alt_color = farbe;}

void tft_push_pop(uint8_t func)
{
  switch(func) {
    case PUSH_ZOOM:
      push_zoom();
      break;
    case POP_ZOOM:
      pop_zoom();
      break;
    case CLR_ZOOM:
      clear_zoom_stack();
      break;
    case PUSH_XY:
      push_xy();
      break;
    case POP_XY:
      pop_xy();
      break;
    case CLR_XY:
      clear_xy_stack();
      break;
    case PUSH_FONT:
      push_font();
      break;
    case POP_FONT:
      pop_font();
      break;
    case CLR_FONT:
      clear_font_stack();
      break;
    case PUSH_COLOR:
      push_color();
      break;
    case POP_COLOR:
      pop_color();
      break;
    case CLR_COLOR:
      clear_color_stack();
      break;
  } // switch()
}    

void tft_init()
{
  tft_font(0);
  tft_fcol(SCHWARZ);
  tft_bcol(WEISS);
  tft_acol(SCHWARZ);
  tft_clear();
  clear_zoom_stack();
  clear_xy_stack();
  clear_font_stack();
  clear_color_stack();
}  

void esc_scanner(char c)            // Verarbeitung von ser. Befehlen
{    
static uint8_t count, bef_len;
short x1=0, y1=0, x2=0, y2=0;

uint8_t temp = 0;
//  printf("%2x  ",c);
  if(count == 0) {        // abhngig vom 1. Befehlszeichen Befehlslnge festsetzen
    switch(c) {             
      case ESC:           // 2 x ESC: * als Sync-Zeichen senden
        bef_len = 0;
        uart0_putc('*');
        esc_flag = 0;
        break;
      case '@':  
        tft_init();
        break;
      case 'C':           // Textcursor setzen
      case 'c':           // diverse 3 Byte Befehle
        bef_len = 3;
        break;
      case 'g':           // Linie, Rechteck leer oder gefuellt
        bef_len = 11;
        break;
      case 'G':           // Grafik: Pixel(x,y) oder Zeichen
        bef_len = 7;
        break;
      case 'V':           // Version und Aufloesung ausgeben
        sprintf(temp_str, "mino PicoTerm-TFT V1.0;%d;%d\0",  MAX_X+1, MAX_Y+1);
        sende_str0(temp_str);
        esc_flag = 0;
        break;
    }   
  }
  if(bef_len == 0) return;          // keine weitere Zeichen
  befehl[count++] = c;              // sonst weitere Zeichen zwischenspeichern
  if(count >= bef_len) {            // und zum Schluss auswerten
    switch (*befehl) {
      case 'C':                     // Textcursor setzen
        zeile = befehl[1];
        spalte = befehl[2];
        tft_set_textcursor(zeile, spalte);
        break;
      
      case 'c':                     // div. Befehle mit drei Zeichen
        temp = befehl[2];
        switch(befehl[1]) {
#ifdef MIT_BLINKEN                        
          case '1':                 // alternative Farbe fuer Blinken
            tft_acol(temp);
            break;
#endif                        
          case '3':
            tft_fcol(temp);         // Schriftfarbe: Vordergrund
            break;
          case '4':
            tft_bcol(temp);         // Schriftfarbe: Hintergrund
            break;
          case '5':                 // Zoom einstellen
            tft_zoom(temp);
            break;
          case '6':                 // Zeichensatz einstellen
            tft_font(temp);
            break;
          case '7':
            tft_f_blink(temp);      // in 10 ms Schritten
            break;
          case '8':
            tft_backlight(temp);    // Abstufung gem Tabelle  
            break;
					case 'b':	                // Baudrate einstellen
            tft_bd_index(temp);
            break;
          case 'O':
            tft_alarm(temp*10);     // Piepdauer in 10 ms Schritten
            break;
          case 't':
            if(temp) touchmeldung = 1;  // automatische Ausgabe per USART
            else touchmeldung = 0;      // keine Ausgabe
            break;
          case 'p':                 // push-pop: zoom, Cursor, Farben
            tft_push_pop(temp);
            break;
        } // case(befehl: c)
          
        case 'g':                   // Linie, Rechteck leer-gefllt
          x1 = (befehl[2]<<8) + befehl[3];
          y1 = (befehl[4]<<8) + befehl[5];
          x2 = (befehl[6]<<8) + befehl[7];
          y2 = (befehl[8]<<8) + befehl[9];
          temp = befehl[10];
          switch(befehl[1]) {
            case 'l':
              zeichne_linie(x1,y1,x2,y2,(temp));
              break;
            case 'r':
              zeichne_rechteck(x1,y1,x2,y2,(temp));
              break;
            case 'f':
              fuelle_rechteck(x1,y1,x2,y2,(temp));
              break;
          }
          break;
        case 'G':                   // Pixel, graf.-Zeichen, Grafikcursor
          x1 = (befehl[2]<<8) + befehl[3];
          y1 = (befehl[4]<<8) + befehl[5];
          temp = befehl[6];
          switch (befehl[1]) {
            case 'p':
              set_pixel(x1,y1,(temp));
              break;
            case 'c':               // Buchstabe pixelgenau ausgeben
              tft_gchar(x1,y1,temp);
              break;
            case 't':               // Grafikcursor setzen
              tft_set_gcursor(x1,y1,temp);
              break;
          }
          break;
      } // switch (*befehl)
    esc_flag = 0;                   // Befehl verarbeitet: alles auf Anfang
    count = 0;
    bef_len = 0;
  }
}


void hauptprogramm()
{
int c;
//  init_t7_touch();
  while(1) {
    do {                            // Eingabe per SPI oder USART abwarten
      if(touch_aktiv) {             // touch-Koordinaten ausgeben, wenn aktiv
        sende_touch();
      }
      c = get_uart0();
    } while(c == EOF);
    if(esc_flag) {
      esc_scanner(c);
      c=0;
    } else {                        // Steuerzeichen
      switch(c) {
        case GLOCKE:
          tft_alarm(300);           // Piepsen
        break;    
        case CLEAR:                 // clear + home_up
          tft_clear();
        case HOME:                  // home-up                        
          akt_x = 0;
          akt_y = 0;
        break;
        case LF:                    // Zeilenvorschub
          akt_y += y_raster;
          if(akt_y + y_raster > MAX_Y+1) {
            scroll();               // bei Bedarf auto-scroll 
            akt_y -= y_raster;      // y nicht erhhen
          }
        break;
        case RTN:
          akt_x = 0;                // an Zeilenanfang
        break;
        case ESC:                   // neuer Befehl
          esc_flag = 1;
        break;
        default:
          tft_ausgabe(c);           // wenn kein Steuerzeichen
      } // switch
    } // if(esc_flag)
  } // while()
}
