/*
Alle Funktionen ohne Gewhr
Michael Nowak
2023-12-02
www.mino-elektronik.de
*/

#define EDT_TFT         // Touch-Leitungen XL und XR sind vertauscht
#define _TOUCH

#include "system_tft_rp2040.h"
#include "touch.h"

#define PIEPSER         20

#define F7_TIMER        7

// ADC-Werte fr Touch-Folien manuell ermittelt, ggf. anpassen
#define X1              70               // temp_x rechts
#define Y1              140
#define X2              940              // temp_x links
#define Y2              825
#define ENTPRELLZEIT    5
#define NACHLAUFZEIT    (ENTPRELLZEIT*2)
static int aktiv = 0, gedrueckt = 0, zuletzt = 0, entprellung = 0;	


static volatile int32_t touch_x=0, touch_y=0;   // Ergebnis der Dekodierung
static volatile int32_t temp_x=0, temp_y=0;
static volatile int32_t x_wert=0, y_wert=0;
static volatile int32_t touch_status;           // Information gedrueckt + repeat

volatile uint8_t sync_flag;

static volatile uint32_t timer = 0;
static volatile short tft_alarm_count, t_blink = 350;
volatile uint8_t blink_bit=0;
volatile uint8_t touch_aktiv;
	

// Zuordnung der Touch-Leitungen
#define XL    26    // = ADC_X
#define YU    27    // = ADC_y
#define XR    22
#define YD    21

//#define ADC_X 26    // ADC0
//#define ADC_Y 27    // ADC1

static inline void x_achse_passiv(void) // TPXL passiv, TPXR passiv
{
  SIO->GPIO_OE_CLR = BIT(XL) |      // XL loeschen
                     BIT(XR);       // XR loeschen
}

static inline void x_achse_0(void)  // TPXR = 0, TPXL = passiv
{
  SIO->GPIO_OE_CLR = BIT(XL);       // XL = Eingang
  SIO->GPIO_OUT_CLR = BIT(XR);      // XR = 0
  SIO->GPIO_OE_SET = BIT(XR);       // XR = Ausgang
}
static inline void x_achse_aktiv(void)  // TPXR = 1, TPXL = 0
{
  SIO->GPIO_OUT_CLR = BIT(XL);      // XL = 0
  SIO->GPIO_OUT_SET = BIT(XR);      // XR = 1
  SIO->GPIO_OE_SET = BIT(XL);       // XL = Ausgang
  SIO->GPIO_OE_SET = BIT(XR);       // XR = Ausgang
}

static inline void x_achse_adc(void)
{
  SIO->GPIO_OE_CLR = BIT(XL) |      // XL loeschen
                     BIT(XR);       // XR loeschen
}

static inline void y_achse_high(void) // TPYD = 1, TPYU = passiv 
{
  SIO->GPIO_OE_CLR = BIT(YU) |      // y-achse passiv
                     BIT(YD);
  SIO->GPIO_OUT_SET = BIT(YD);      // YD = 1
  SIO->GPIO_OE_SET = BIT(YD);       // YD = Ausgang
}

static inline void pullup_YD_ein(void)
{
  PADS_BANK0_SET->GPIO12 = PADS_BANK0_GPIO0_PUE_Msk;  // Pullup einschalten
}

static inline void pullup_YD_aus(void)
{
  PADS_BANK0_CLR->GPIO12 = PADS_BANK0_GPIO0_PUE_Msk;  // Pullup wieder abschalten
}

static inline void y_achse_aktiv(void)  // TPYD = 1, TPYU = 0
{
  SIO->GPIO_OUT_CLR = BIT(YU);      // YU = 0
  SIO->GPIO_OUT_SET = BIT(YD);      // YD = 1
  SIO->GPIO_OE_SET = BIT(YU) |      // y-achse aktiv
                     BIT(YD);
}

static inline void y_achse_adc(void)  // TPYD + TPYU passiv
{
  SIO->GPIO_OE_CLR = BIT(YU) |      // y-achse aktiv
                     BIT(YD);
}

static inline int32_t lese_adc_x(void)
{
  ADC->CS = ADC_CS_START_ONCE_Msk | // Einzelmessung
            0<<ADC_CS_AINSEL_Pos |  // ADC0
            ADC_CS_EN_Msk;          // starten
  while(!(ADC->CS & ADC_CS_READY_Msk));
  return(ADC->RESULT/4);            // nur 10 Bit
}	

static inline int32_t lese_adc_y(void)
{
  ADC->CS = ADC_CS_START_ONCE_Msk | // Einzelmessung
            1<<ADC_CS_AINSEL_Pos |  // ADC1
            ADC_CS_EN_Msk;          // starten
  while(!(ADC->CS & ADC_CS_READY_Msk));
  return(ADC->RESULT/4);            // nur 10 Bit
}

void dekodiere_touch(void)
{
static int phase, nachlauf;
	phase++;				                  // immer erhhen
	switch(phase) {
		case 1:				                  // x auf schwebend und y auf V+ schalten
      x_achse_passiv();
			y_achse_high();
			break;
		case 2:				                  // y auf ADC und sehen, wo der Pegel bleibt
      x_achse_0();
      pullup_YD_ein();              // y-achse auf '1' halten
			y_achse_adc();  
      break;
		case 3:				                  // y-wert messen
      if(lese_adc_y() < 150) {      // touch ist gedrueckt
        aktiv = 1;
        nachlauf = 0;
			} else {
        aktiv = 0;
        nachlauf++;
        if(nachlauf >= NACHLAUFZEIT) {
          nachlauf--;
          phase--;                  // auf Kontakt vom touch warten
        } 
      }  
			break;
		case 4:				                  // x_achse einschalten
      pullup_YD_aus();              // pullup wegnehmen
			y_achse_adc();
			x_achse_aktiv();
			break;
    case 5:				                  // y_achse einlesen
			x_wert = temp_x = lese_adc_y();
			x_achse_adc();
			y_achse_aktiv();
			break;
		case 6:				                  // x-achse einlesen
			y_wert = temp_y = lese_adc_x();
			phase = 0;                    // wieder bei case1 beginnen
			break;
		default:
			phase = 0;		                // fr ungltige Zustnde	
	}	

	if(phase == 0) {			
		if(aktiv != zuletzt) {
			entprellung = 0;
			zuletzt = aktiv;
		}			
		entprellung++;
		if(entprellung > ENTPRELLZEIT) {
			entprellung--;
			if(aktiv ^ gedrueckt) {		    // nderung passiert
				gedrueckt = aktiv;
				if(gedrueckt) {             // skalieren
					touch_x = ((temp_x-X1)*SIZE_X)/ (X2-X1);
					touch_y = ((temp_y-Y1)*SIZE_Y)/ (Y2-Y1);
					if(touch_x > MAX_X) touch_x = MAX_X;
					if(touch_x < 0) touch_x = 0;
					if(touch_y > MAX_Y) touch_y = MAX_Y;
					if(touch_y < 0) touch_y = 0;
#ifdef EDT_TFT
          touch_x = SIZE_X-touch_x; // vertauschte Leitungen
#endif                    
					touch_aktiv = TOUCH_AKTIV;// touch ist gedrckt
				}
				else {
					touch_aktiv = TOUCH_PASSIV; // touch wurde losgelassen
				}
			}
		}	
	}
}

void teste_touch_status(void)                        // 
{
static int change,last_change;
static int touch_aktiv_alt,rep_counter=0, rep_flag=0;

  if(rep_counter) rep_counter--;

  change = (touch_aktiv_alt ^ touch_aktiv);
  touch_aktiv_alt = touch_aktiv;
  if(change) {
    last_change = change;
    rep_counter = FIRST_REPEAT;
    rep_flag = 0;
  } else if(rep_counter == 0 && rep_flag) {
    change = last_change & touch_aktiv;
    rep_counter = NEXT_REPEAT;
  }
  if(change) {
    if(((change & TOUCH_AKTIV) & touch_aktiv) == TOUCH_AKTIV) {
      if(rep_flag) {
        touch_status = REPEAT_TASTE + TOUCH_GEDRUECKT;
      } else {
        touch_status = NEUE_TASTE + TOUCH_GEDRUECKT;
      }
      rep_flag = 1;
    } else {
      rep_flag = 0;
    }
  }
}


// ADC einschalten
static void adc12_init(void)
{
  RESETS_CLR->RESET = RESETS_RESET_adc_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_adc_Msk));
}

void init_t7_touch(void)
{
  RESETS_CLR->RESET = RESETS_RESET_pwm_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pwm_Msk));
  IO_BANK0->GPIO21_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // YD
  IO_BANK0->GPIO22_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // XR
  IO_BANK0->GPIO26_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // XL
  IO_BANK0->GPIO27_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // YU
  PADS_BANK0_CLR->GPIO21 = PADS_BANK0_GPIO0_PDE_Msk;          // Pulldown abschalten
  PADS_BANK0_CLR->GPIO22 = PADS_BANK0_GPIO0_PDE_Msk;          // Pulldown abschalten
  PADS_BANK0_CLR->GPIO26 = PADS_BANK0_GPIO0_PDE_Msk;          // Pulldown abschalten
  PADS_BANK0_CLR->GPIO27 = PADS_BANK0_GPIO0_PDE_Msk;          // Pulldown abschalten

  IO_BANK0->GPIO25_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // LED-Ausgang
  SIO->GPIO_OE_SET = 1 << LOCAL_LED; 
  
  PWM->CH7_DIV = 100 << PWM_CH0_DIV_INT_Pos;      // Vorteiler / 100
  PWM->CH7_TOP = SystemCoreClock/100/(1000-1);    // 1 kHz
  NVIC_EnableIRQ(PWM_IRQ_WRAP_IRQn);
  NVIC_SetPriority(PWM_IRQ_WRAP_IRQn, PRIO_LOW);
  
  PWM->CH7_CSR =  1;                              // enable ms-timer
  PWM->INTE =  1<<F7_TIMER;                       // enable ISR

  adc12_init();
}
  


void lese_touch(int *t_x, int *t_y)
{
  PWM_CLR->INTE = 1<<F7_TIMER;      // Interrupt sperren
  __ISB();                          // Ausfuehrung abwarten
  *t_x = touch_x;
  *t_y = touch_y;
  PWM_SET->INTE = 1<<F7_TIMER;      // enable ISR
}



int get_tft_touch(int *x, int *y)   // fr direkte Verwendung
{
int temp = touch_status;
  if(temp) touch_status = 0;
  lese_touch(x,y);
  return(temp);	
}

    
void schreibe_piepser(uint8_t ein_aus)
{
  ein_aus = (ein_aus ^ 1) & 1;      // invertierte Ausgabe
  IO_BANK0->GPIO20_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // PIEPSER-Ausgang
  SIO->GPIO_OE_SET = BIT(PIEPSER); 
  if(ein_aus) SIO->GPIO_OUT_SET = BIT(PIEPSER); 
  else SIO->GPIO_OUT_CLR = BIT(PIEPSER); 
}    

void tft_f_blink(uint8_t wert)
{
	if(wert == 255) wert = 0;	        // 255 oder 0: kein blinken
	t_blink = wert * 10;	            // auf 10 ms-Schritte bringen
}	

void set_timer(uint32_t wert)
{
	timer = wert;
}

uint32_t get_timer()
{
	return(timer);
}	

void sync()                         // alle 1 ms von TIM7
{
	sync_flag = 0;
	while(!sync_flag);
}

void wait(int16_t zeit)
{
	while(zeit-- > 0) sync();
}

void tft_alarm(int16_t wert)
{
	schreibe_piepser(1);              // einschalten
	tft_alarm_count = wert;
}			
  
void blinken_ein()
{ blink_bit = 1;}

void blinken_aus()
{ blink_bit = 0;}

// 1 kHz Aufrufe fr Messzeiten
void PWM_IRQ_WRAP_IRQHandler(void)
{
static uint16_t blink_ctr;
  if(PWM->INTS & 1<<F7_TIMER) {     // PWM_CH7?
    PWM->INTR = 1<<F7_TIMER;        // clear flag

    timer++;
    sync_flag = 1;
    if(tft_alarm_count) {
      tft_alarm_count--;
      if(tft_alarm_count == 0) 
        schreibe_piepser(0);	      // abschalten
    }	

    if(t_blink) {	
      blink_ctr++;
      if(blink_ctr >= t_blink) {
        blink_ctr = 0;
        SIO->GPIO_OUT_XOR = BIT(LOCAL_LED); // zur Kontrolle
        blink_bit ^= 1;
      }	
    } else {
        blink_bit = 0;	            // kein Blinken
    }       
    dekodiere_touch();
    teste_touch_status();
  }
}
