#define _USART0_PICO

#define RX_TX_INVERS                // Invertierung von RS232-Treiber umgehen

/*
2023-12-10
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#include "system_tft_rp2040.h"

#define UART0_BUF_LEN  	  0x400
#define UART0_BUF_MASK  	(UART0_BUF_LEN-1)


static uint8_t uart0_rx_puffer[UART0_BUF_LEN];
static uint32_t uart0_rx_schreib_zeiger,
                uart0_rx_lese_zeiger;

void UART0_IRQ_IRQHandler(void)
{
uint32_t temp, sr_temp;
uint32_t temp_zeiger;
  sr_temp = UART0->UARTMIS;
  if(sr_temp & UART0_UARTMIS_RXMIS_Msk) {         // Rx-Flag testen
    temp = UART0->UARTDR;                         // Zeichen lesen
    temp_zeiger = (uart0_rx_schreib_zeiger + 1) & UART0_BUF_MASK;
    if( temp_zeiger != uart0_rx_lese_zeiger) {
      uart0_rx_puffer[uart0_rx_schreib_zeiger] = temp;
      uart0_rx_schreib_zeiger = temp_zeiger;
    }
    UART0_SET->UARTICR = UART0_UARTICR_RXIC_Msk;  // flag lschen
  }
}

void clr_uart0_buf()
{
  UART0_CLR->UARTIMSC = UART0_UARTIMSC_RXIM_Msk;  // RXIE abschalten
  __ISB();
  uart0_rx_lese_zeiger = 0;
  uart0_rx_schreib_zeiger = 0;
  UART0_SET->UARTIMSC = UART0_UARTIMSC_RXIM_Msk;  // RXIE einschalten
}

int read_uart0_buf()
{
int32_t temp = 0;                                 // default bei leerem Puffer
  if(uart0_rx_lese_zeiger != uart0_rx_schreib_zeiger) {
    temp = uart0_rx_puffer[uart0_rx_lese_zeiger++];
    uart0_rx_lese_zeiger &= UART0_BUF_MASK;
  }
  return(temp);
}

uint32_t read_uart0_buf_count()
{
	return((uart0_rx_schreib_zeiger - uart0_rx_lese_zeiger) & UART0_BUF_MASK);
}	

void set_baudrate0(uint32_t baud)
{
  baud = (SystemCoreClock << 2) / baud;
  UART0->UARTIBRD = baud / 64;
  UART0->UARTFBRD = baud % 64;
  UART0->UARTLCR_H = 3 << UART0_UARTLCR_H_WLEN_Pos; // RxD+TxD 8-bit  
}

// TX = GPIO16
// RX = GPIO17
void init_uart0(void)
{
  RESETS_CLR->RESET = RESETS_RESET_uart0_Msk | 
                      RESETS_RESET_usbctrl_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_uart0_Msk));

  IO_BANK0->GPIO16_CTRL = IO_BANK0_GPIO16_CTRL_FUNCSEL_uart0_tx;
  IO_BANK0->GPIO17_CTRL = IO_BANK0_GPIO17_CTRL_FUNCSEL_uart0_rx;
  
#ifdef RX_TX_INVERS       // fuer Test ohne MAX232
  IO_BANK0->GPIO16_CTRL = IO_BANK0_GPIO16_CTRL_FUNCSEL_uart0_tx |
                         1 << IO_BANK0_GPIO0_CTRL_OUTOVER_Pos;
  IO_BANK0->GPIO17_CTRL = IO_BANK0_GPIO17_CTRL_FUNCSEL_uart0_rx |
                         1 <<  IO_BANK0_GPIO1_CTRL_INOVER_Pos;
#endif

  set_baudrate0(DEF_BAUDRATE);                    // Baudrate einstellen

  NVIC_EnableIRQ(UART0_IRQ_IRQn);
  NVIC_SetPriority(UART0_IRQ_IRQn,PRIO_MID_HIGH);
  
  UART0_SET->UARTIMSC = UART0_UARTIMSC_RXIM_Msk;  // Rx ISR zulassen
  UART0_SET->UARTCR = UART0_UARTCR_UARTEN_Msk;    // und UART0 aktivieren
  clr_uart0_buf();
}	

void uart0_putc(int zeichen)
{
  while (UART0->UARTFR & UART0_UARTFR_TXFF_Msk);
  UART0->UARTDR = zeichen;
}

int putchar(int x)                  // stdout fuer printf()
{
  if((uint8_t)x == 10) uart0_putc(13); 
  uart0_putc(x);
  return(x);
}

int get_uart0(void)
{
  if(read_uart0_buf_count()) {
    return(read_uart0_buf());
  } else {
    return(EOF);                    // bei leerem Fifo -1
  }
}  

void sende_str0(char *s)            // ser. Ausgabe
{
  while(*s) uart0_putc(*(s++));
  uart0_putc(13);                   // CR
  uart0_putc(10);                   // LF
} 
